/*
 * This file is part of the 'DEX online' code base (www.dexonline.ro).
 * Author: Catalin Francu
 * 
 * Included in Dixit with trivial modifications.
 * Modified by: Octavian Procopiuc <oprocopiuc@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You can read the GNU General Public License at
 * http://www.gnu.org/licenses/licenses.html
 */


#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "htmlize.h"

#define NUM_SUFFIXES 9
char suffixes[NUM_SUFFIXES][2][10] =
  { { "iei", "ie" }, { "ului", "" }, { "ul", "" }, { "uri", "" },
    { "urilor", "" }, { "ilor", "" }, { "i", "" }, { "~a", "" }, { "e", "" } };

/*************************** Trie implementation *****************************/

typedef struct _trie {
  char* payload;
  struct _trie* ptr[256];
} trie;

trie* int2uni = NULL;

trie* trie_make() {
  trie* t = new trie;
  t->payload = NULL;
  memset(t->ptr, 0, sizeof(t->ptr));
  return t;
}

void trie_put(trie* t, char* str, char* payload) {
  for (char* s = str; *s; s++) {
    unsigned char c = (unsigned char)*s;
    if (!t->ptr[c]) t->ptr[c] = trie_make();
    t = t->ptr[c];
  }
  if (t->payload)
    fprintf(stderr, "%s already associated with payload %s\n",
            str, t->payload);
  t->payload = strdup(payload);
}

// matches the shortest possible prefix of str
// modifies *len to indicate the length of the prefix that matched.
char* trie_get(trie* t, char* str, int* len) {
  *len = 0;
  while (*str && t && !t->payload) {
    t = t->ptr[(unsigned char)*str];
    str++;
    (*len)++;
  }
  return t ? t->payload : NULL;
}

/* [tavi] Changed these to UTF-8 codes. */
void create_int2uni_trie() {
  int2uni = trie_make();
  /*trie_put(int2uni, " - ", " \342\200\223 ");*/
  trie_put(int2uni, " ** ", " \342\231\246 ");
  trie_put(int2uni, " * ", " \342\227\212 ");
  trie_put(int2uni, "'A", "\303\201");
  trie_put(int2uni, "'a", "\303\241");
  trie_put(int2uni, "`A", "\303\200");
  trie_put(int2uni, "`a", "\303\240");
  trie_put(int2uni, ":A", "\303\204");
  trie_put(int2uni, ":a", "\303\244");
  trie_put(int2uni, "~A", "\304\202");
  trie_put(int2uni, "~a", "\304\203");
  trie_put(int2uni, "^A", "\303\202");
  trie_put(int2uni, "^a", "\303\242");
  trie_put(int2uni, "'^A", "'\303\202");
  trie_put(int2uni, "'^a", "'\303\242");
  trie_put(int2uni, "^'A", "'\303\202");
  trie_put(int2uni, "^'a", "'\303\242");
  trie_put(int2uni, "'~A", "'\304\202");
  trie_put(int2uni, "'~a", "'\304\203");
  trie_put(int2uni, "~'A", "'\304\202");
  trie_put(int2uni, "~'a", "'\304\203");
  trie_put(int2uni, "\\C", "\303\207");
  trie_put(int2uni, "\\c", "\303\247");
  trie_put(int2uni, "'C", "\304\206");
  trie_put(int2uni, "'c", "\304\207");
  trie_put(int2uni, "~C", "\304\214");
  trie_put(int2uni, "~c", "\304\215");
  trie_put(int2uni, "`E", "\303\210");
  trie_put(int2uni, "`e", "\303\250");
  trie_put(int2uni, "'E", "\303\211");
  trie_put(int2uni, "'e", "\303\251");
  trie_put(int2uni, "^E", "\303\212");
  trie_put(int2uni, "^e", "\303\252");
  trie_put(int2uni, ":E", "\303\213");
  trie_put(int2uni, ":e", "\303\253");
  trie_put(int2uni, "~E", "\304\224");
  trie_put(int2uni, "~e", "\304\225");
  trie_put(int2uni, "~G", "\304\236");
  trie_put(int2uni, "~g", "\304\237");
  trie_put(int2uni, "'^I", "'\303\216");
  trie_put(int2uni, "'^i", "'\303\256");
  trie_put(int2uni, "^'I", "'\303\216");
  trie_put(int2uni, "^'i", "'\303\256");
  trie_put(int2uni, "`I", "\303\214");
  trie_put(int2uni, "`i", "\303\254");
  trie_put(int2uni, "'I", "\303\215");
  trie_put(int2uni, "'i", "\303\255");
  trie_put(int2uni, "^I", "\303\216");
  trie_put(int2uni, "^i", "\303\256");
  trie_put(int2uni, ":I", "\303\217");
  trie_put(int2uni, ":i", "\303\257");
  trie_put(int2uni, "~I", "\304\254");
  trie_put(int2uni, "~i", "\304\255");
  /*trie_put(int2uni, "~N", "&#xd1;");*/
  trie_put(int2uni, "~n", "\303\261");
  trie_put(int2uni, "`O", "\303\222");
  trie_put(int2uni, "`o", "\303\262");
  trie_put(int2uni, "'O", "\303\223");
  trie_put(int2uni, "'o", "\303\263");
  trie_put(int2uni, "^O", "\303\224");
  trie_put(int2uni, "^o", "\303\264");
  trie_put(int2uni, ":O", "\303\226");
  trie_put(int2uni, ":o", "\303\266");
  /*trie_put(int2uni, "~R", "&#x158;");*/
  trie_put(int2uni, "~r", "\305\231");
  trie_put(int2uni, "~S", "\305\240");
  trie_put(int2uni, "~s", "\305\241");
  trie_put(int2uni, "\\S", "\305\236");
  trie_put(int2uni, "\\s", "\305\237");
  trie_put(int2uni, "\\T", "\305\242");
  trie_put(int2uni, "\\t", "\305\243");
  trie_put(int2uni, "`U", "\303\231");
  trie_put(int2uni, "`u", "\303\271");
  trie_put(int2uni, "'U", "\303\232");
  trie_put(int2uni, "'u", "\303\272");
  trie_put(int2uni, "^U", "\303\233");
  trie_put(int2uni, "^u", "\303\273");
  trie_put(int2uni, ":U", "\303\234");
  trie_put(int2uni, ":u", "\303\274");
  /*trie_put(int2uni, "~U", "\305\254");*/
  trie_put(int2uni, "~u", "\305\255");
  trie_put(int2uni, "'Y", "\303\235");
  trie_put(int2uni, "'y", "\303\275");
  trie_put(int2uni, ":y", "\303\277");
  /*trie_put(int2uni, "~Z", "\305\275");*/
  trie_put(int2uni, "~z", "\305\276");

  trie_put(int2uni, "&#xab;", "\302\253");
  trie_put(int2uni, "&#xbb;", "\302\273");
  trie_put(int2uni, "&#x2260;", "\342\211\240");
  trie_put(int2uni, "&#x7e;", "~");
  trie_put(int2uni, "&#x2a;", "*");
  trie_put(int2uni, "&#x25;", "%");
  trie_put(int2uni, "&#xb0;", "\302\260");
  trie_put(int2uni, "&deg;", "\302\260");
  trie_put(int2uni, "&#x201e;", "\"");
  trie_put(int2uni, "&#x201d;", "\"");
  trie_put(int2uni, "&#x5B;", "[");
  trie_put(int2uni, "&#x5D;", "]");
}

// From |text|ref|, generate a HTML reference
void generate_html_reference(char* text, int len_text,
			     char* ref, int len_ref,
			     char* html) {
  // Empty references: copy the text as it is
  if (ref[0] == '\0') {
    strcpy(ref, text);
    len_ref = len_text;
  }

  // "-" references: copy the text and remove the suffix
  if (ref[0] == '-' && ref[1] == '\0') {
    strcpy(ref, text);
    len_ref = len_text;
    bool found = false;
    for (int i = 0; !found && i < NUM_SUFFIXES; i++) {
      int len_suf = strlen(suffixes[i][0]);
      if (len_ref >= len_suf &&
	  !strcmp(ref + len_ref - len_suf, suffixes[i][0])) {
	strcpy(ref + len_ref - len_suf, suffixes[i][1]);
	found = true;
      }
    }
  }

  // Only keep the last word in a multiple word reference. But be careful,
  // in a reference like "heraldica (@1@)" you want to keep "heraldica".
  // Ergo, use the last word (space-delimited) that contains at least one
  // letter.
  char* ref_begin = NULL;
  bool in_word = false;
  char* s;
  char* p;

  for (s = ref; *s; s++) {
    if (isalpha(*s) ||
	((*s == '~' || *s == '\\' || *s == '^') && isalpha(*(s + 1)))) {
      if (!in_word) ref_begin = s;
      in_word = true;
    } else {
      in_word = false;
      *s = '\0';
    }
  }

  if (!ref_begin) ref_begin = ref;
  // [tavi] Remove non-letter chars.
  for (s = ref_begin, p = ref_begin; *s; s++) {
    if (isalpha(*s))
      *p++ = tolower(*s);
  }
  *p = '\0';

  sprintf(html, "<a href=\"%s\">%s</a>",
	  ref_begin, text);
}

void insert_explicit_references(char* olds, char* news) {
  char text[100], ref[100], html[200];
  int len_news = 0, len_text = 0, len_ref = 0;
  int mode = 0;     // 0 = not between bars; 1 = text; 2 = reference

  for (char* s = olds; *s; s++) {
    if (*s == '|') {
      switch (mode) {
      case 0:
	break; // Nothing really
      case 1:
	text[len_text] = '\0';
	break;
      case 2:
	// Here we actually append the reference
	ref[len_ref] = '\0';
	generate_html_reference(text, len_text, ref, len_ref, html);
	int len_html = strlen(html);
	strcpy(news + len_news, html);
	len_news += len_html;
	text[0] = '\0'; len_text = 0;
	ref[0] = '\0'; len_ref = 0;
	break;
      }
      mode = (mode + 1) % 3;
    } else {
      switch (mode) {
      case 0: news[len_news++] = *s; break;
      case 1: text[len_text++] = *s; break;
      case 2: ref[len_ref++] = *s;
      }
    }
  }

  news[len_news] = '\0';
}

// All sorts of substitutions, mostly for diacritics, but also for dashes
// and quotes.
void insert_diacritics(char* olds, char* news) {
  int len = 0;
  bool in_tag = false, in_bold = false, in_italic = false, in_quotes = false;
  bool in_spaced = false, first_in_spaced = false;

  // In spaced-out mode, do not insert &nbsp; between & and ;. This would
  // cause %&#x103;% to be printed as & # x 1 0 3 ;, which would mess up
  // the HTML.
  bool in_amp_spaced = false;

  char* s = olds;
  while (*s) {
    if (*s == '<') in_tag = true;
    else if (*s == '>') in_tag = false;

    if (in_tag) {
      // Don't touch ANYTHING between < and >
      news[len++] = *s++;
    } else if (*s == '"') {
      in_quotes = !in_quotes;
#if defined(WIN32) || defined(_WIN32) || defined(__WIN32__) || defined(__NT__)
      const char* r = "\"";
#else
      const char* r = in_quotes ? "\342\200\236" : "\342\200\235";
#endif
      /*in_quotes ? "&#x201e;" : "&#x201d;";*/
      strcpy(news + len, r);
      len += strlen(r);
      s++;
    } else if (*s == '@') {
      in_bold = !in_bold;
      const char* r = in_bold ? "<b>" : "</b>";
      strcpy(news + len, r);
      len += strlen(r);
      s++;
    } else if (*s == '$') {
      in_italic = !in_italic;
      const char* r = in_italic ? "<i>" : "</i>";
      strcpy(news + len, r);
      len += strlen(r);
      s++;
    } else if (*s == '%') {
      // Only toggle in_spaced. Then, before each character below, add &nbsp;
      in_spaced = !in_spaced;
      first_in_spaced = true; // No &nbsp; before the first character
      s++;
    } else {
      // Add spacing if we're between spaces
      if (in_spaced) {
	if (!first_in_spaced && !in_amp_spaced && *s != ',' &&
	    *s != '.' && *s != ' ' && *(s-1) != '^') {
	  const char* r = "&nbsp;";
	  strcpy(news + len, r);
	  len += strlen(r);
	}
	first_in_spaced = false;

	if (*s == '&') in_amp_spaced = true;
	else if (*s == ';') in_amp_spaced = false;
      }

      // Match a pattern beginning at this char, or just copy one char.
      int match_len;
      char* payload = trie_get(int2uni, s, &match_len);
      if (payload) {
        strcpy(news + len, payload);
        len += strlen(payload);
        s += match_len;
      } else {
        news[len++] = *s++;
      }
    }
  }
  news[len] = '\0';
}

// Subsitutes *only* the diacritics present in the table. Leaves references
// untouched.
void insert_diacritics_only(char* olds, char* news) {
  int len = 0;
  char* s = olds;
  int bar_mod = 0; // Count bars modulo 3

  while (*s) {
    if (*s == '|')
      bar_mod = (bar_mod + 1) % 3;

    bool found = false;
    if (bar_mod != 2) {
      // Try to search for a matching pattern beginning at this char.
      int match_len;
      char* payload = trie_get(int2uni, s, &match_len);
      if (payload) {
        strcpy(news + len, payload);
        len += strlen(payload);
        s += match_len;
        found = true;
      }
    }
    
    // No diacritic here
    if (!found) news[len++] = *s++;
  }
  news[len] = '\0';
}

// And subscripts too! :)
void insert_superscripts(char* olds, char* news) {
  int len = 0;
  char* s = olds;
  char buf[29];

  while (*s) {
    if (*s == '^' || *s == '_') {
      char* end = s + 1;
      while (isdigit(*end) || (*end == '+') ||
	     (*end == '-' && end == s + 1)) end++;
      int num_digits = end - s - 1;
      if (num_digits) {
	// [tavi] Windows complains about non-constant value in initializer.
	//char buf[num_digits + 20];
	strcpy(buf, (*s == '^') ? "<sup>" : "<sub>");
	strncat(buf, s + 1, num_digits);
	strcat(buf, (*s == '^') ? "</sup>" : "</sub>");
	int buf_len = strlen(buf);
	strcpy(news + len, buf);
	len += buf_len;
	s = end;
      } else {
	news[len++] = *s++;
      }
    } else {
      news[len++] = *s++;
    }
  }
  news[len] = '\0';
}

char* my_gets(char* s, int len) {
  fgets(s, len, stdin);
  int slen = strlen(s);

  while (slen && isspace(s[slen -1]))
    s[--slen] = '\0';

  char* begin = s;
  while (isspace(*begin)) begin++;
  return begin;
}

void htmlize(char* out, const char* in) {
  if (int2uni == NULL)
    create_int2uni_trie();

  // We will use two arrays. At each operation we make on the string, we will
  // read data from one and store the result into the other.
  char s1[50000], s2[50000];

  strncpy(s1, in, 50000);
  insert_explicit_references(s1, s2);
  insert_diacritics(s2, s1);
  insert_superscripts(s1, s2);
  strncpy(out, s2, 50000);
}

